#!/usr/bin/env python

"""
Module to handle tracking the closest points to some target.

Used by segysqldist.py.
    
     Copyright (C) 2005  Kurt Schwehr

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

$Id: closestN.py,v 1.5 2005/06/23 02:55:33 schwehr Exp $

"""

import mmap
import os
import struct	# Convert two and from binary formats
import sqlite
import sys
import math

#from distance import greatCircleDistKm
def greatCircleDistKm(lonlat1, lonlat2):
    # FIX:  BADDDD!  This is degree distance
    a1 = float(lonlat1[1]); b1 = float(lonlat1[0])
    a2 = float(lonlat2[1]); b2 = float(lonlat2[0])

    return math.sqrt((a1-a2)*(a1-a2) + (b1-b2)*(b1-b2))



def printPoints(pts):
    for p in pts:
        for value in p:
            print "  %.3f" % value,
        print ';    ',
    print
    return

def printPointsGnuplot(pts):
    for p in pts:
        for value in p:
            print "  %.6f" % value,
        print
    return

class ClosestPoints:
    """
    Handle keeping track of the closest N points to some location

    Throw points at this class and it will keep track of the best
    """
    def __init__(self,lonlat, N=100):
        """
        Setup a closest point object.  Distance is in km.

        lonlat - location of the target point to compare with (decimal degrees)
        N      - specifies the number of points to keep
        """
        self.N = N
        self.points = []     # list of lon,lat,distance tuples
        self.pointData = []  # Save user supplied data with points
        self.bestDist = 999999999
        self.worstDist = self.bestDist + 1

        self.location = lonlat

        return

    def getInsertPoint(self,distance):
        """ Find the index where this distance should be inserted into the list"""
        if len(self.points)==0: return 0
        index = None
        for i in range(0,len(self.points)):#self.N):
            pointDist = self.points[i][2]
            if distance < pointDist:
                index = i  # Found it
                break
        if index == None and len(self.points)<self.N: index = len(self.points)
        if index != None: assert(index>=0)
        return index 

    def almostEqual(self,a,b,delta=0.0001):
        #print a,b,b-delta, "    ", a,b,b+delta
        if (a<(b-delta)) or (a>(b+delta)): return False
        return True

    def isInPoints(self,lonlat):
        """ Super slow idiot way to figure this out"""
        for p in self.points:
            if self.almostEqual(lonlat[0],p[0]) and self.almostEqual(lonlat[1],p[1]):
                return True
        return False
    
    def checkPoint(self,lonlat, data=None, ignore=False):
        """ Check a point to compare against what is already known
        
        lonlat - decimal degrees of the point to compare
        data   - any data you want stored with point
        ignore - set to true to not insert points that have the same location
        	for data sets that have many pings with the same latlon
        """
        dist = greatCircleDistKm(lonlat,self.location)
        #print "Currentdistance: ", dist, lonlat, self.location
        triplet = (lonlat[0],lonlat[1],dist)
        if len(self.points)<self.N:
            index = self.getInsertPoint(dist)
            assert(None != index)
            
            if not ignore and self.isInPoints(lonlat):
                #print "Skipping point with same location"
                return
            
            self.points.insert(index,triplet)
            self.pointData.insert(index,data)
        else:
            index = self.getInsertPoint(dist)
            if index != None:
                if not ignore and self.isInPoints(lonlat):
                    #print "Skipping point with same location.  List is full"
                    return
                index = self.getInsertPoint(dist)
                self.points.insert(index,triplet)
                self.pointData.insert(index,data)
                # Drop worst point off the list
                self.points.pop(-1)
                self.pointData.pop(-1)
        return


######################################################################
# Temporary test code

from optparse import OptionParser
from segy_version import VERSION
myparser = OptionParser(usage="%prog [options] segyfile1 segyfile2 ... segyfileN",
                      version="%prog "+'999'+' - '+'666')
myparser.add_option('-f','--file',dest='dbfilename',default='segy.db',
		    help='file for the sqlite database  [default: %default]')
myparser.add_option('-x','--x',dest='x', help='x coordinate (decimal degrees longitude)')
myparser.add_option('-y','--y',dest='y', help='y coordinate (decimal degrees latitude)')

if __name__ == "__main__":
    (options,args) = myparser.parse_args()
    #print options
    if not options.x or not options.y:
        print "You must specify both x and y\n"
        myparser.print_help()
        sys.exit("Goodbye and good ridance")

    target = (options.x,options.y)

    o = open ("point","w")
    o.write(options.x+" "+options.y+"\n")

    pointList = [
        (0,0),
        (1,1),
        (10,10),
        (1,1.5)
        ]
    
    
    ########################################
    # N points
    ########################################
    if True:
        print "# Beginning search"
        cPts = ClosestPoints(target,2)
        i = 0
        o = open ('data','w')
        for row in pointList:
            print "\n# ",i, " len: ", len(cPts.points)
            i += 1
            #if i > 1000:
            #    print "FIX: early exit"
            #    break 
            x = row[0]
            y = row[1]
            o.write(str(x)+' '+str(y)+'\n')
            point = (float(x),float(y))
            data = ("a","b")
            cPts.checkPoint(point,data)
        o = open ('results','w')

        for p in cPts.points:
            s = ""
            for value in p:
                snew = s + ("  %.6f" % value)
                s = snew
            s += '\n'
            o.write(s)


