#!/usr/bin/env python
"""
Find the closest points to a target location using the trace header database.

     Copyright (C) 2005  Kurt Schwehr

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

$Id: segysqldist.py,v 1.12 2005/10/05 16:51:03 schwehr Exp $

"""

# These values end up in the pydoc web page
__version__ = '$Revision: 1.12 $' # See man ident
__date__ = '$Date: 2005/10/05 16:51:03 $' # FIX: pull out just the date
__author__ = 'Kurt Schwehr'
__credits__ = """Lots of people for helping me with my thesis.
Especially Neal Driscoll and Lisa Tauxe!
"""

import mmap
import os
import struct	# Convert two and from binary formats
import sqlite
import sys


BOMBASTIC= 4
VERBOSE  = 3
TRACE    = 2
TERSE    = 1
ALWAYS   = 0


from segysqlgmt import CoordTransform

from optparse import OptionParser

# What version do we want to track?  This would be for cvs of this file
vcvs = "$Revision: 1.12 $"
vcvs = vcvs[vcvs.find(':')+2:]
vcvs = vcvs[:vcvs.find('$')-1]
version_cvs = vcvs
del (vcvs) # stop polution


from segy_version import VERSION

myparser = OptionParser(usage="%prog [options] segyfile1 segyfile2 ... segyfileN",
                      version="%prog "+VERSION+' - '+version_cvs)
myparser.add_option('-f','--file',dest='dbfilename',default='segy.db',
		    help='file for the sqlite database  [default: %default]')

myparser.add_option('-X','--longitude',dest='x', type='float', default=None,
                    help='x coordinate (decimal degrees longitude)')
myparser.add_option('-Y','--latitude',dest='y', type='float', default=None,
                    help='y coordinate (decimal degrees latitude)')

myparser.add_option('--in-position-file',dest='inPosFile',default=None,
                    help='load the x y lon lat from a file instead of from -X and -Y')


myparser.add_option('-n','--num-shots',dest='N', default='1000', help='number of shots to find [default: %default].  More is slower')

myparser.add_option('-p','--point-file',dest='pointFile', default='point.dat', help='file to write out target x,y [default: %default].')

myparser.add_option('-d','--data-file',dest='dataFile', default=None, help='file to write out a subset of data searched [default: %default].')
myparser.add_option('-s','--sample-interval',dest='sampleInterval', default=1000, help='how many samples to skip when writing out --data-file [default: %default].')


myparser.add_option('-o','--output-file',dest='resultsFile', default='results-lonlat.dat', help='file to write out the closest points to [default: %default].')

myparser.add_option('-S','--do-not-skip',dest='ignoreSame', default=False, action='store_true',help='give this flag to keep samples in the list with the same xy values.  Some systems give the same GPS position over a number of shots.')

myparser.add_option('-c','--coord-unit',type="int",dest='forceCoordUnit',default=-1,
                    help='Force a SEGY coordinate system, e.g. 2 == seconds of arc.  Use this for short format databases. [default: follow file]')


myparser.add_option('-v',action="count",dest='verbosity',default=0,
                    help='how much information to give.  Specify multiple times to increase verbosity')
myparser.add_option('--verbosity',dest='verbosity',type='int',
                    help='Specify verbosity.  Should be in the range of '
                    +str(ALWAYS)+'...'+str(BOMBASTIC)+' (None...Bombastic)')

myparser.add_option('-x','--flip-x',dest='flipX',default=False,action='store_true',
                    help='Multiply the x coordinate by -1')
myparser.add_option('-y','--flip-y',dest='flipY',default=False,action='store_true',
                    help='Multiply the x coordinate by -1')


myparser.add_option('-H','--force-hemisphere',type="int",dest='forceHemisphere',default=-1,
                    help='Force hemisphere for shots that have bad navigation.'+
                    '  0-NW,  1-NE,  2-SW,  3-SE')

myparser.add_option('-u','--from-utm',action="store_true",dest='fromUTM',default=False,
                    help='Assume x and y are in UTM and convert to Long/Lat')
myparser.add_option('-z','--utm-zone',type="int",dest='utmZone',default=None,
                    help='Specify the UTM zone if converting from UTM to Long/Lat')


myparser.add_option('-Z','--skip-zeros',default=False,dest='skipZeros',action='store_true',
                    help='Skip points that have 0 0 stored in the x,y fields')


myparser.add_option('-m','--min-file-key',dest='minFileKey',default=None, type='int',
                    help='Set this to start processing at a certain file index [default: %default]')
myparser.add_option('-M','--max-file-key',dest='maxFileKey',default=None, type='int',
                    help='Set this to end processing at a certain file index [default: %default]')

myparser.add_option('-k','--key',dest='key',default='FieldRecNo', type='string',
                    help='An additional key that you would like to see for each close match [default: %default]')

myparser.add_option('--navigation-file',dest='navFile',default=None, type='string',
                    help='Ignore the database X and Y fields.  Read from a shotpoint lon lat file.  e.g. segysqlshiptrack.py -i --dump-key Shotpoint')


from closestN import ClosestPoints
    
if __name__ == "__main__":
    (options,args) = myparser.parse_args()


    if 0 != len(args):
        print 'extra args: ',args
        sys.exit( "ERROR: there should be no extra args\n")
        # FIX: force no extra args... fail out

    if options.inPosFile != None:
        if None != options.x or None != options.y:
            sys.exit("ERROR: must only specify (-X and -Y) or position file, not both")
        x,y = open(options.inPosFile,'r').readline().split()
        options.x, options.y = float(x),float(y)

    if not options.x or not options.y:
        print "You must specify both x and y\n"
        myparser.print_help()
        sys.exit("Goodbye and good ridance")

    if options.verbosity >= TRACE:
        print 'Searching for best match to ',options.x, options.y

    if options.x<-360. or options.x>360:
        sys.exit("ERROR: Longitude must be in the range of [-360...360]")
    if options.y<-90.  or options.y> 90:
        sys.exit("ERROR: Latitude  must be in the range of [ -90... 90]")

    if options.navFile == None:
        options.navData=None
    else:
        options.navData={}
        for line in open (options.navFile,'r'):
            if 0==len(line) or '#'== line[0]: continue # Skip blanks or comments
            fields = line.split()
            shot,x,y = int(fields[0]),float(fields[1]),float(fields[2])
            options.navData[shot] = (x,y)
    if options.verbosity >= BOMBASTIC*2:
        for key in options.navData.keys():
            print key,options.navData[key]

    dbName = options.dbfilename
    if not os.access(dbName,os.R_OK):
        print "\nERROR: database does not exist...",dbName
        sys.exit("  No shoes, no shirt, no service")
    cx = sqlite.connect (dbName)
    ct = CoordTransform(cx,options)
    cu = cx.cursor()



    print "Starting SQL select.  This make take quite a while"

    from segysqldump import hasField
    hasEnsembleNo = hasField(cx,'EnsembleNo','segyTrace')

    # FIX: make sure ensemble no is in the database
    searchStr = 'SELECT X,Y,traceNumber,fileKey,Shotpoint,'
    if hasEnsembleNo: searchStr += ' EnsembleNo, '
    searchStr += options.key+' FROM segyTrace '
    if (None != options.minFileKey) or (None != options.maxFileKey): searchStr += ' WHERE'
    if (None != options.minFileKey): searchStr += ' fileKey>='+str(options.minFileKey)
    if (None != options.minFileKey) and (None != options.maxFileKey): searchStr += ' and'
    if (None != options.maxFileKey): searchStr += ' fileKey<='+str(options.maxFileKey)
    searchStr += ';'

    if options.verbosity>=VERBOSE: print searchStr
    cu.execute(searchStr)
    print "Finished SQL select"
    target = (options.x,options.y)

    o = open (options.pointFile,"w")
    o.write(str(options.x)+" "+str(options.y)+"\n")
    o.close()
    del(o)

    ########################################
    # N points
    ########################################
    i = 0
    if None != options.dataFile:
        o = open (options.dataFile,'w')
    print "# Beginning search"
    if 1<int(options.N):
        print "Using slower search for many points.  Suggest using -n 1 for faster"
        cPts = ClosestPoints(target,int (options.N))
        print "Inspecting data:"
        for row in cu.fetchall():
            x,y = None,None

            if options.navData != None:
                try:
                    pt = options.navData[int(row.Shotpoint)]
                    x,y = pt[0],pt[1]
                except KeyError:
                    sys.exit('ERROR: Shotpoint '+str(row.Shotpoint)+' not in the supplied nav file!')
                    # Could alternatively just warn and fall back on sql database
            else:
                x,y = ct.getLonLat(row['x'],row['y'])

            i += 1
            if 0== i % 1000: print i,x,y

            if None != options.dataFile:
                o.write(str(x)+' '+str(y)+'\n')  # FIX: maybe writeout a subset?
            point = (float(x),float(y))
            if hasEnsembleNo:
                data = (row['fileKey'],row['Shotpoint'],row['traceNumber'],row['EnsembleNo'],row[options.key])
            else:
                data = (row['fileKey'],row['Shotpoint'],row['traceNumber'],row[options.key])
            cPts.checkPoint(point,data,ignore=options.ignoreSame)
        if None != options.dataFile: del(o)

        o = open (options.resultsFile,'w')
        
        o.write('# segysqldist.py\n')
        o.write('# \n')
        o.write('# distance    - currently is in "degrees"... sort of\n')
        o.write('# traceNumber - the offset in the particular segy file\n')
        o.write('# fileKey     - use the segyFile table in *.db to translate\n')
        o.write('# \n')
        o.write('# longitude latitude "distance" fileKey Shotpoint traceNumber '+options.key+'\n')
        for i in range(len(cPts.points)):
            p = cPts.points[i]
            d = cPts.pointData[i]
            s = ''
            for value in p:
                s += ("  %.6f" % value)
            for value in d:
                s += " " + str(value)
            s += '\n'
            o.write(s)
        del(o)

    ########################################
    # Single point - Much faster
    ########################################
    else:
        assert(1==int(options.N))
        bestFile = None
        bestShot = None
        bestX = None
        bestY = None
        bestDistance = 99999999999.
        bestEnsembleNo = None
        bestKey = None

        # FIX: use a real distance function!!!!
        from closestN import greatCircleDistKm

        for row in cu.fetchall():

            if options.navData != None:
                try:
                    pt = options.navData[int(row.Shotpoint)]
                    x,y = pt[0],pt[1]
                except KeyError:
                    sys.exit('ERROR: Shotpoint '+str(row.Shotpoint)+' not in the supplied nav file!')
                    # Could alternatively just warn and fall back on sql database
            else:
                x,y = ct.getLonLat(row['x'],row['y'])

            i += 1

            if None != options.dataFile:
                o.write(str(x)+' '+str(y)+'\n')  # FIX: maybe writeout a subset?
            dist = greatCircleDistKm((x,y),target)

            if dist > 1000:
                print "WARNING: core farther than 1000 KM.  Are you missing a minus sign?"

            if 0== i % 1000: print i,x,y,dist
            
            if dist < bestDistance:
                #print x,y,row['traceNumber'], row['fileKey'],row['Shotpoint'],target,dist
                bestDistance = dist
                bestShot = row['Shotpoint']
                bestX = x
                bestY = y
                bestFile = row['fileKey']
                if hasEnsembleNo:
                    bestEnsembleNo = row['EnsembleNo']
                else:
                    bestEnsembleNo = ''
                bestKey = row[options.key]
        if options.verbosity>=TRACE:
            print 'BEST fields: fileKey, shotpoint, x, y, distance, EnsembleNo,'+options.key
        print "BEST: ",bestFile,bestShot,bestX,bestY,bestDistance,bestEnsembleNo,bestKey
        o = open (options.resultsFile,'w')
        o.write(str(bestX)+' '+str(bestY)+' '+str(bestDistance)+'\n')
        del(o)
    print "Thanks for playing"
